#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @author Paul Dechamps
# @date 2024
# Test the blaster adjoint implementation.

# Imports.

import blast.utils as vutils
import blast
import numpy as np

from fwk.wutils import parseargs
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

import fwk
from fwk.testing import *
from fwk.coloring import ccolors
import numpy as np

def cfgInviscid(nthrds, verb):
    import os.path
    # Parameters
    return {
    # Options
    'scenario' : 'aerodynamic',
    'task' : 'optimization',
    'Threads' : nthrds, # number of threads
    'Verb' : verb, # verbosity
    # Model (geometry or mesh)
    'File' : os.path.dirname(os.path.abspath(__file__)) + '/../../models/dart/n0012.geo', # Input file containing the model
    'Pars' : {'xLgt' : 5, 'yLgt' : 5, 'growthRatio' : 2., 'msTe' : 0.1, 'msLe' : 0.1}, # parameters for input file model
    'Dim' : 2, # problem dimension
    'Format' : 'gmsh', # save format (vtk or gmsh)
    # Markers
    'Fluid' : 'field', # name of physical group containing the fluid
    'Farfield' : ['upstream', 'farfield', 'downstream'], # LIST of names of physical groups containing the farfield boundaries (upstream/downstream should be first/last element)
    'Wing' : 'airfoil', # LIST of names of physical groups containing the lifting surface boundary
    'Wake' : 'wake', # LIST of names of physical group containing the wake
    'WakeTip' : 'wakeTip', # LIST of names of physical group containing the edge of the wake
    'Te' : 'te', # LIST of names of physical group containing the trailing edge
    'Upstream' : 'upstream',
    # Freestream
    'M_inf' : 0.2, # freestream Mach number
    'AoA' : 2., # freestream angle of attack
    # Geometry
    'S_ref' : 1., # reference surface length
    'c_ref' : 1., # reference chord length
    'x_ref' : .25, # reference point for moment computation (x)
    'y_ref' : 0.0, # reference point for moment computation (y)
    'z_ref' : 0.0, # reference point for moment computation (z)
    # Numerical
    'LSolver' : 'SparseLU', # inner solver (Pardiso, MUMPS or GMRES)
    'G_fill' : 2, # fill-in factor for GMRES preconditioner
    'G_tol' : 1e-5, # tolerance for GMRES
    'G_restart' : 50, # restart for GMRES
    'Rel_tol' : 1e-6, # relative tolerance on solver residual
    'Abs_tol' : 1e-8, # absolute tolerance on solver residual
    'Max_it' : 20, # solver maximum number of iterations
    }

def cfgBlast(verb):
    return {
        'Re' : 1e7,                     # Freestream Reynolds number
        'Minf' : 0.2,                   # Freestream Mach number (used for the computation of the time step only)
        'CFL0' : 1,                     # Inital CFL number of the calculation
        'Verb': verb,                   # Verbosity level of the solver
        'couplIter': 100,               # Maximum number of iterations
        'couplTol' : 1e-10,             # Tolerance of the VII methodology
        'iterPrint': 20,                # int, number of iterations between outputs
        'resetInv' : True,              # bool, flag to reset the inviscid calculation at every iteration.
        'sections' : [0],               # List of sections for boundary layer calculation
        'xtrF' : [0.2, 0.2],            # Forced transition locations
        'interpolator' : 'Matching',    # Interpolator for the coupling
    }

def main():
    # Timer.
    tms = fwk.Timers()
    tms['total'].start()

    args = parseargs()
    icfg = cfgInviscid(args.k, args.verb)
    vcfg = cfgBlast(args.verb)

    alpha = icfg['AoA']*np.pi/180
    dalpha = 1e-6 * np.pi/180
    daoa = [alpha - dalpha, alpha + dalpha]

    coupler, isol, vsol = vutils.initBlast(icfg, vcfg)
    morpher = isol.iobj['mrf']

    # Adjoint objects
    cAdj = blast.CoupledAdjoint(isol.adjointSolver, vsol)

    # Run coupled case
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['forward'].start()
    aeroCoeffs = coupler.run()
    tms['forward'].stop()

    print(ccolors.ANSI_BLUE + 'PyDartAdjoint...' + ccolors.ANSI_RESET)
    tms['adj_dart'].start()
    isol.adjointSolver.run()
    tms['adj_dart'].stop()

    print(ccolors.ANSI_BLUE + 'PyBlasterAdjoint...' + ccolors.ANSI_RESET)
    tms['adj_blast'].start()
    cAdj.run()
    tms['adj_blast'].stop()

    # Get mesh matrix
    dCl_x = np.zeros((isol.iobj['bnd'].nodes.size(), isol.getnDim()))
    dCd_x = np.zeros((isol.iobj['bnd'].nodes.size(), isol.getnDim()))
    for inod, n in enumerate(isol.blw[0].nodes):
        for i in range(isol.getnDim()):
            dCl_x[inod, i] = cAdj.tdCl_x[n.row][i]
            dCd_x[inod, i] = cAdj.tdCd_x[n.row][i]

    print(ccolors.ANSI_BLUE + 'PyFiniteDifference...' + ccolors.ANSI_RESET)
    # Finite difference on AoA
    tms['fd_blaster_aoa'].start()
    dcl = []
    dcd = []
    for aa in daoa:
        isol.setAoA(aa)
        coupler.reset()
        aeroCoeffs = coupler.run(write=False)
        dcl.append(isol.getCl())
        dcd.append(isol.getCd()+vsol.Cdf)
    dCl_AoA_FD = (dcl[-1] - dcl[-2])/(2.*dalpha)
    dCd_AoA_FD = (dcd[-1] - dcd[-2])/(2.*dalpha)
    tms['fd_blaster_aoa'].stop()
    isol.setAoA(icfg['AoA']*np.pi/180) # Reset AoA after FD

    # Recover grad AoA from grad x
    drot = np.array([[-np.sin(alpha), np.cos(alpha)], [-np.cos(alpha), -np.sin(alpha)]])

    dCl_AoA_fromX = 0.
    dCd_AoA_fromX = 0.
    for inod, n in enumerate(isol.iobj['bnd'].nodes):
        dx = drot.dot([n.pos[0] - 0.25, n.pos[1]])
        dCl_AoA_fromX += np.array([cAdj.tdCl_x[n.row][0], cAdj.tdCl_x[n.row][1]]).dot(dx)
        dCd_AoA_fromX += np.array([cAdj.tdCd_x[n.row][0], cAdj.tdCd_x[n.row][1]]).dot(dx)
    
    ### BLASTER FD MESH ###
    tms['fd_blaster_msh'].start()
    saveNodes = np.zeros((len(isol.iobj['pbl'].msh.nodes), isol.getnDim()))
    for n in isol.iobj['pbl'].msh.nodes:
        for i in range(isol.getnDim()):
            saveNodes[n.row, i] = n.pos[i]
    dCl_x_blaster_fd = np.zeros((len(isol.blw[0].nodes), isol.getnDim()))
    dCd_x_blaster_fd = np.zeros((len(isol.blw[0].nodes), isol.getnDim()))
    delta = 1e-6
    cmp = 0
    import time
    start_time = time.time()
    for inod, n in enumerate(isol.blw[0].nodes):
        # Safety check
        if isol.iobj['bnd'].nodes[inod].row != n.row:
            raise RuntimeError('Nodes do not match', n.row, isol.iobj['bnd'].nodes[inod].row)
        for idim in range(isol.getnDim()):
            cl = [0, 0]
            cd = [0, 0]
            for isgn, sign in enumerate([-1, 1]):
                morpher.savePos()
                n.pos[idim] = saveNodes[n.row, idim] + sign * delta
                morpher.deform()
                isol.updateMesh()
                
                coupler.reset()
                # prevent print
                from contextlib import redirect_stdout
                with redirect_stdout(None):
                    coupler.run(write=False)
                cmp += 1
                if cmp % 20 == 0:
                    print('F-D opers', cmp, '/', isol.blw[0].nodes.size() * isol.getnDim() * 2, '(node '+str(n.row)+')', 'time', str(round(time.time() - start_time, 3))+'s')
                cl[isgn] = isol.getCl()
                cd[isgn] = isol.getCd() + vsol.Cdf

                # Reset mesh
                for node in isol.iobj['sol'].pbl.msh.nodes:
                    for d in range(isol.getnDim()):
                        node.pos[d] = saveNodes[node.row, d]
            dCl_x_blaster_fd[inod, idim] = (cl[1] - cl[0]) / (2. * delta)
            dCd_x_blaster_fd[inod, idim] = (cd[1] - cd[0]) / (2. * delta)
    tms['fd_blaster_msh'].stop()

    print('Statistics')
    print(tms)
    print(ccolors.ANSI_BLUE + 'PyTesting' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('dCl_dAoA', cAdj.tdCl_AoA, dCl_AoA_FD, 1e-5))
    tests.add(CTest('dCd_dAoA', cAdj.tdCd_AoA, dCd_AoA_FD, 1e-4))
    tests.add(CTest('norm dCl_dx (adj - fd)', np.linalg.norm(dCl_x), np.linalg.norm(dCl_x_blaster_fd), 1e-6))
    tests.add(CTest('norm dCd_dx (adj - fd)', np.linalg.norm(dCd_x), np.linalg.norm(dCd_x_blaster_fd), 1e-4))
    tests.add(CTest('dCl_dAoA (msh)', dCl_AoA_fromX, cAdj.tdCl_AoA, 1e-2)) # The value gets closer if the mesh if refined
    tests.add(CTest('dCd_dAoA (msh)', dCd_AoA_fromX, cAdj.tdCd_AoA, 1e-1)) # The value gets closer if the mesh if refined
    tests.run()

    # eof
    print('')

if __name__ == '__main__':
    main()